from typing import Any, List, Sequence, Set, Type, Union

import polars
import pyarrow
from pydantic import BaseModel
from typing_extensions import TypeAlias

from chalk.utils import AnyDataclass


class StreamResolverParam(BaseModel):
    name: str


class StreamResolverParamMessage(StreamResolverParam):
    typ: Union[Type[str], Type[bytes], Type[BaseModel], AnyDataclass]


StreamResolverWindowType: TypeAlias = Union[
    Type[List[str]],
    Type[List[bytes]],
    Type[List[BaseModel]],
    Type[AnyDataclass],
    Type[pyarrow.Table],
    Type[polars.DataFrame],
    Any,  # The annotation value is likely going to be a GenericAlias, which messes with pydantic's validation
]


class StreamResolverParamMessageWindow(StreamResolverParam):
    typ: StreamResolverWindowType


class StreamResolverSignature(BaseModel):
    params: Sequence[StreamResolverParam]
    output_feature_fqns: Set[str]


class StreamResolverParamKeyedState(StreamResolverParam):
    typ: Union[Type[BaseModel], Type[AnyDataclass]]
    default_value: Any



from os import PathLike
from typing import Any, Dict, TypeVar, Union

from sqlalchemy.engine.url import URL

from chalk.client.client_impl import ChalkConfigurationException
from chalk.sql.base.sql_source import BaseSQLSource, TableIngestMixIn

T = TypeVar("T")


class SQLiteInMemorySourceImpl(BaseSQLSource, TableIngestMixIn):
    def __init__(self):
        super(SQLiteInMemorySourceImpl, self).__init__()
        self.ingested_tables: Dict[str, Any] = {}

    def local_engine_url(self) -> Union[str, URL]:
        return "sqlite:///:memory:?check_same_thread=true"


class SQLiteFileSourceImpl(BaseSQLSource, TableIngestMixIn):
    def __init__(self, filename: Union[PathLike, str]):
        try:
            import aiosqlite
        except ModuleNotFoundError:
            raise ChalkConfigurationException.missing_dependency("chalkpy[sqlite]")
        del aiosqlite
        self.filename = filename
        self.ingested_tables: Dict[str, Any] = {}
        super(SQLiteFileSourceImpl, self).__init__()

    def local_engine_url(self) -> Union[str, URL]:
        return f"sqlite:///{str(self.filename)}?check_same_thread=true"



"""
Adds thread-local context to a Python logger. Taken from neocrym/log-with-context
"""
import contextvars
import logging
from typing import Any, Callable, Dict, Optional, Union

_EXTRA_TYPE = Dict[str, Any]

_LOGGING_CONTEXT: contextvars.ContextVar[_EXTRA_TYPE] = contextvars.ContextVar("_LOGGING_CONTEXT")


def init_extra():
    """Initialize our thread-local variable for storing contexts."""
    if _LOGGING_CONTEXT.get(None) is None:
        _LOGGING_CONTEXT.set({})


def get_extra() -> _EXTRA_TYPE:
    """
    Retrieve the thread-local extra.
    This initializes the thread-local variable if necessary.
    """
    init_extra()
    return _LOGGING_CONTEXT.get()


def set_extra(extra: _EXTRA_TYPE) -> _EXTRA_TYPE:
    """
    Sets the thread-local context to a new value.
    This erases whatever the context used to be. If you would rather
    restore that old value later, you might prefer using the
    :py:class:`add_logging_context` context manager.
    """
    _LOGGING_CONTEXT.set(extra)
    return extra


class Logger:
    """
    Wrapper for a python :py:class:`logging.Logger`.
    This wrapper reads the current local context and emits
    them for each log message.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        logger: Optional[Union[logging.Logger, "Logger"]] = None,
    ):
        """
        Initialize.
        Args:
            logger: A logging.Logger instance to wrap.
        """
        if isinstance(logger, Logger):
            self.base_logger: Any = logger.base_logger
        else:
            self.base_logger = logger or logging.getLogger(name=name)

    def _msg(self, func: Callable, msg, *args, **kwargs):
        """Log with our extra values,"""
        extra = {
            **self.extra,
            **kwargs.pop("extra", {}),
        }
        # Because we wrap a logging.Logger instance through 2 layers
        # of redirection, we need to add 2 to the logger's stacklevel
        # so we correctly log the logging statement's line number and function name
        original_stacklevel = kwargs.pop("stacklevel", 1)
        stacklevel = original_stacklevel + 2
        return func(msg, *args, extra=extra, stacklevel=stacklevel, **kwargs)

    @property
    def extra(self) -> _EXTRA_TYPE:
        """Return the extra metadata that this logger sends with every message."""
        return get_extra()

    def debug(self, msg, *args, **kwargs):
        """Debug."""
        return self._msg(self.base_logger.debug, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """Info."""
        return self._msg(self.base_logger.info, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """Warning."""
        return self._msg(self.base_logger.warning, msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        """Warn. Deprecated. Use `warning` instead."""
        return self._msg(self.base_logger.warn, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """Error."""
        return self._msg(self.base_logger.error, msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """Critical."""
        return self._msg(self.base_logger.critical, msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """Log."""
        return self._msg(self.base_logger.log, level, msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        """Exception."""
        return self._msg(self.base_logger.exception, msg, *args, **kwargs)


class add_logging_context:
    """
    A context manager to push and pop "extra" dictionary keys.
    """

    def __init__(self, **kwargs):
        """Create a new context manager."""
        self._new_extra = kwargs
        self._old_extra = {}

    def __enter__(self):
        """Add the new kwargs to the thread-local state."""
        self._old_extra = get_extra()
        set_extra({**self._old_extra, **self._new_extra})
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Return the thread-local state to what it used to be."""
        set_extra(self._old_extra)
        return False



from typing import cast

from chalk.features import features
from chalk.sql import PostgreSQLSource
from chalk.sql.integrations.postgres import PostgreSQLSourceImpl


@features
class UserFeatures:
    id: str


@features
class LibraryFeatures:
    id: str


def test_table_builder():
    source: PostgreSQLSourceImpl = cast(
        PostgreSQLSourceImpl,
        PostgreSQLSource()
        .with_table(name="users", features=UserFeatures)
        .with_table(name="libraries", features=LibraryFeatures),
    )
    assert len(source.ingested_tables) == 2



import re
from typing import Any, Callable, Dict, List, Mapping, Optional, Set, Tuple, Union

from mypy.nodes import (
    ARG_NAMED_OPT,
    ARG_OPT,
    ARG_POS,
    GDEF,
    Argument,
    AssignmentStmt,
    Block,
    CallExpr,
    ClassDef,
    Expression,
    JsonDict,
    NameExpr,
    PlaceholderNode,
    RefExpr,
    SymbolTable,
    SymbolTableNode,
    TypeInfo,
    Var,
)
from mypy.plugin import (
    AnalyzeTypeContext,
    ClassDefContext,
    DynamicClassDefContext,
    FunctionContext,
    MethodContext,
    MypyFile,
    Plugin,
    SemanticAnalyzerPluginInterface,
    lookup_fully_qualified,
)
from mypy.plugins.common import add_method, deserialize_and_fixup_type
from mypy.types import Instance, NoneType, Type, UnboundType

from chalk.features import DataFrame, Features, feature, features
from chalk.features.dataframe import DataFrameMeta


def _get_full_name(cls: Union[type, Callable]):
    return f"{cls.__module__}.{cls.__name__}"


class PluginKeywords:
    @staticmethod
    def fullname_is_dataframe_class(fullname: str):
        return re.match(r"chalk\.feature_n\.feature_[0-9]+\.feature.DataFrame", fullname) is not None

    feature_set_decorators = {_get_full_name(features)}

    dataframe_meta = _get_full_name(DataFrameMeta)

    feature_makers = {
        _get_full_name(feature),
    }

    features_class = _get_full_name(Features)

    dataframe_class = _get_full_name(DataFrame)

    @staticmethod
    def get_features_from_metadata(metadata: Optional[Mapping[str, JsonDict]]) -> List[str]:
        return metadata.get("features", {}).get("attributes", []) if metadata is not None else []

    @staticmethod
    def set_features_metadata(obj, cols: List[Any]) -> None:
        obj.metadata["features"] = {"attributes": cols}

    @staticmethod
    def get_dataframe_cols_from_metadata(metadata: Mapping[str, JsonDict]) -> List[str]:
        return metadata.get("dataframe", {}).get("columns", [])

    @staticmethod
    def set_dataframe_cols_metadata(obj, cols: List[str]) -> None:
        obj.metadata["dataframe"] = {"columns": cols}

    @staticmethod
    def get_dataframe_class_fullname(n: int) -> str:
        return f"chalk.feature_n.feature_{str(n)}.feature.DataFrame"

    @staticmethod
    def get_feature_class_fullname(n: int) -> str:
        return f"chalk.feature_n.feature_{str(n)}.feature.Features"

    any_typevar: str = f"chalk.feature_n.feature_2.feature.T1"


class FeatureWrapper:
    def __init__(
        self,
        name: str,
        feature_name: str,
        line: int,
        column: int,
        type: Optional[Type],
        info: TypeInfo,
    ):
        self.feature_name = feature_name
        self.name = name
        self.line = line
        self.column = column
        self.type = type
        self.info = info

    def to_argument(self) -> Argument:
        return Argument(
            variable=self.to_var(),
            type_annotation=self.type,
            initializer=None,
            kind=ARG_NAMED_OPT,
        )

    def to_var(self) -> Var:
        return Var(self.name, self.type)

    def serialize(self) -> JsonDict:
        assert self.type
        return {
            "name": self.name,
            "feature_name": self.feature_name,
            "line": self.line,
            "column": self.column,
            "fullname": self.info.fullname,
            "type": self.type.serialize(),
        }

    @classmethod
    def deserialize(cls, info: TypeInfo, data: JsonDict, api: SemanticAnalyzerPluginInterface) -> "FeatureWrapper":
        data = data.copy()
        if data.get("kw_only") is None:
            data["kw_only"] = False
        typ = deserialize_and_fixup_type(data.pop("type"), api)
        return cls(type=typ, info=info, **data)


def _collect_field_args(expr: Expression, ctx: ClassDefContext) -> Tuple[bool, Dict[str, Expression]]:
    """Returns a tuple where the first value represents whether
    the expression is a call to feature and the second is a
    dictionary of the keyword arguments that feature() was called with.
    """
    if (
        isinstance(expr, CallExpr)
        and isinstance(expr.callee, RefExpr)
        and expr.callee.fullname in PluginKeywords.feature_makers
    ):
        # feature() only takes keyword arguments.
        args = {}
        for name, arg, kind in zip(expr.arg_names, expr.args, expr.arg_kinds):
            if not kind.is_named():
                if kind.is_named(star=True):
                    # This means that `field` is used with `**` unpacking,
                    # the best we can do for now is not to fail.
                    # TODO: we can infer what's inside `**` and try to collect it.
                    message = 'Unpacking **kwargs in "feature()" is not supported'
                else:
                    message = '"feature()" does not accept positional arguments'
                ctx.api.fail(message, expr)
                return True, {}
            assert name is not None
            args[name] = arg
        return True, args
    return False, {}


def collect_attributes(ctx: ClassDefContext) -> Optional[List[FeatureWrapper]]:
    """Collect all attributes declared in the dataclass and its parents.

    All assignments of the form

      a: SomeType
      b: SomeOtherType = ...

    are collected.
    """
    # First, collect attributes belonging to the current class.
    cls = ctx.cls
    attrs: List[FeatureWrapper] = []
    known_attrs: Set[str] = set()
    for stmt in cls.defs.body:
        # Any assignment that doesn't use the new type declaration
        # syntax can be ignored out of hand.
        if not (isinstance(stmt, AssignmentStmt) and stmt.new_syntax):
            continue

        # a: int, b: str = 1, 'foo' is not supported syntax so we
        # don't have to worry about it.
        lhs = stmt.lvalues[0]
        if not isinstance(lhs, NameExpr):
            continue

        sym = cls.info.names.get(lhs.name)
        if sym is None:
            # This name is likely blocked by a star import. We don't need to defer because
            # defer() is already called by mark_incomplete().
            continue

        node = sym.node
        if isinstance(node, PlaceholderNode):
            # This node is not ready yet.
            return None
        assert isinstance(node, Var)

        # x: ClassVar[int] is ignored by dataclasses.
        if not hasattr(node, "is_classvar"):
            continue
        if node.is_classvar:
            continue

        # has_field_call, field_args = _collect_field_args(stmt.rvalue, ctx)

        known_attrs.add(lhs.name)

        attrs.append(
            FeatureWrapper(
                name=lhs.name,
                feature_name=lhs.name,
                line=stmt.line,
                column=stmt.column,
                type=sym.type,
                info=cls.info,
            )
        )

    return attrs


def features_decorator_callback(ctx: ClassDefContext) -> None:
    info = ctx.cls.info
    attributes = collect_attributes(ctx)
    if attributes is None:
        # Some definitions are not ready, defer() should be already called.
        return
    for attr in attributes:
        if attr.type is None:
            ctx.api.defer()
            return

    if ("__init__" not in info.names or info.names["__init__"].plugin_generated) and attributes:
        args = [attr.to_argument() for attr in attributes]

        if info.fallback_to_any:
            # Make positional args optional since we don't know their order.
            # This will at least allow us to typecheck them if they are called
            # as kwargs
            for arg in args:
                if arg.kind == ARG_POS:
                    arg.kind = ARG_OPT

            # nameless_var = Var('')
            # args = [
            #    Argument(nameless_var, AnyType(TypeOfAny.explicit), None, ARG_STAR),
            #    *args,
            #    Argument(nameless_var, AnyType(TypeOfAny.explicit), None, ARG_STAR2),
            # ]

        add_method(
            ctx,
            "__init__",
            args=args,
            return_type=NoneType(),
        )

    for attr in attributes:
        class_name = attr.feature_name.replace(".", "__")
        class_def = ClassDef(class_name, Block([]))
        fullname = ctx.api.qualified_name(class_name)
        class_def.fullname = fullname

        attr_info = TypeInfo(SymbolTable(), class_def, ctx.api.cur_mod_id)
        class_def.info = attr_info
        if not hasattr(attr.type, "type"):
            continue
        attr_info.mro = [attr_info, attr.type.type]
        # meta = ctx.api.named_type('chalk.features.FeatureComparisonMeta')
        # attr_info.declared_metaclass = meta
        attr.info = attr_info
        if getattrs(attr, "type.type.declared_metaclass.type.fullname") == PluginKeywords.dataframe_meta:
            PluginKeywords.set_dataframe_cols_metadata(
                obj=attr_info,
                cols=[
                    a.name if isinstance(a, UnboundType) else a.type.fullname for a in getattr(attr.type, "args", [])
                ],
            )

        # Guide here: https://github.com/python/mypy/blob/5bd2641ab53d4261b78a5f8f09c8d1e71ed3f14a/test-data/unit/plugins/dyn_class_from_method.py
        table_node = SymbolTableNode(kind=GDEF, node=attr_info, plugin_generated=True)
        ctx.api.add_symbol_table_node(attr.name, table_node)

    PluginKeywords.set_features_metadata(info, [attr.serialize() for attr in attributes])

    return None


def features_variadic_generic_callback(ctx: AnalyzeTypeContext) -> Type:
    args = ctx.type.args
    arg_types = [ctx.api.analyze_type(arg) for arg in args]
    if len(arg_types) == 1:
        return arg_types[0]

    if len(arg_types) == 0:
        return ctx.type

    return ctx.api.named_type(
        PluginKeywords.get_feature_class_fullname(len(arg_types)),
        sorted(arg_types, key=lambda x: x.type.name if hasattr(x, "type") else ""),
    )


def init_features_callback(ctx: FunctionContext) -> Type:
    return_type = ctx.default_return_type.type

    kwarg_names = [arg[0] for arg in ctx.arg_names if isinstance(arg, list) and len(arg) == 1]

    symbol_table_params = sorted([return_type.names.get(k) for k in kwarg_names], key=lambda x: x.fullname)

    if len(symbol_table_params) == 1:
        return Instance(symbol_table_params[0].node, args=[])

    if len(kwarg_names) == 0:
        return ctx.default_return_type

    feature_wrapper = lookup_fully_qualified(
        name=PluginKeywords.get_feature_class_fullname(len(kwarg_names)),
        modules=ctx.api.modules,
    )

    args = []
    for tp in symbol_table_params:
        node = tp.node
        if isinstance(tp.node, Var):
            node = node.info
        if isinstance(node, TypeInfo):
            s = Instance(node, args=[])
            args.append(s)

    return Instance(
        typ=feature_wrapper.node,
        args=args,
    )


def add_attributes_to_dataframe(ctx: AnalyzeTypeContext) -> Type:
    args = ctx.type.args
    arg_types = [ctx.api.analyze_type(arg) for arg in args]

    override_types = []
    for a in arg_types:
        features = PluginKeywords.get_features_from_metadata(getattrs(a, "type.metadata") or {})
        if len(features) > 0:
            for f in features:
                override_types.append(UnboundType(f["fullname"]))

    if len(override_types) > 0:
        return ctx.api.named_type(
            PluginKeywords.get_dataframe_class_fullname(len(override_types)),
            sorted(override_types, key=lambda x: x.name),
        )

    if len(args) > 0:
        return ctx.api.named_type(
            PluginKeywords.get_dataframe_class_fullname(len(arg_types)),
            sorted(arg_types, key=lambda x: getattrs(x, "type.name") or ""),
        )

    return ctx.type


def narrow_dataframe_maker(
    parent_members_fullnames: List[str],
) -> Callable[[AnalyzeTypeContext], Type]:
    def dataframe_callback(ctx: AnalyzeTypeContext) -> Type:
        args = [a for a in ctx.type.args if getattrs(a, "base_type_name") != "typing.Any"]
        filtered_members = [ctx.api.analyze_type(arg) for arg in args]
        filtered_members_not_in_parent = set([f.type.fullname for f in filtered_members]) - set(
            parent_members_fullnames
        )

        def format_name(names: List[str]):
            return ", ".join(sorted([".".join([a for a in n.split(".") if a][-2:]) for n in names]))

        if len(filtered_members_not_in_parent) > 0:
            ctx.api.fail(
                f"""Could not narrow DataFrame. Missing {format_name(list(filtered_members_not_in_parent))}
Parent:         {format_name(parent_members_fullnames)}
Child:          {format_name([f.type.fullname for f in filtered_members])}
Child - Parent: {format_name(list(filtered_members_not_in_parent))}""",
                ctx.type,
            )

        if len(filtered_members) == 0:
            return ctx.type

        named_type = ctx.api.named_type(
            PluginKeywords.get_dataframe_class_fullname(len(filtered_members)),
            sorted(filtered_members, key=lambda x: x.type.fullname),
        )

        return named_type

    return dataframe_callback


def getattrs(obj: Any, path: str):
    current = obj
    for p in path.split("."):
        current = getattr(current, p, None)
    return current


class ChalkPlugin(Plugin):
    def get_type_analyze_hook(self, fullname: str) -> Optional[Callable[[AnalyzeTypeContext], Type]]:
        if PluginKeywords.features_class == fullname:
            return features_variadic_generic_callback

        if PluginKeywords.dataframe_class == fullname:
            return add_attributes_to_dataframe

        if PluginKeywords.fullname_is_dataframe_class(fullname):
            return add_attributes_to_dataframe

        # The property we're accessing is a dataframe
        fully_qualified = self.lookup_fully_qualified(fullname)

        if fully_qualified is not None:
            dataclass_type_params = PluginKeywords.get_dataframe_cols_from_metadata(
                getattrs(fully_qualified, "node.metadata") or {}
            )
            if len(dataclass_type_params) > 0:
                return narrow_dataframe_maker(dataclass_type_params)

        return None

    def get_method_hook(self, fullname: str) -> Optional[Callable[[MethodContext], Type]]:
        if fullname.startswith("chalk.sql.integrations") and fullname.endswith("query"):
            # TODO: Support chalk query
            return None
        return None

    def get_dynamic_class_hook(self, fullname: str) -> Optional[Callable[[DynamicClassDefContext], None]]:
        return None

    def get_function_hook(self, fullname: str) -> Optional[Callable[[FunctionContext], Type]]:
        fully_qualified = self.lookup_fully_qualified(fullname)
        node_metadata = (getattrs(fully_qualified, "node.metadata") or {}).get("features", {}).get("attributes")

        if node_metadata is not None:
            return init_features_callback

        return None

    def get_class_decorator_hook(self, fullname: str) -> Optional[Callable[[ClassDefContext], None]]:
        if fullname in PluginKeywords.feature_set_decorators:
            return features_decorator_callback

        return None

    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        fullname = file.fullname
        assert isinstance(fullname, str)
        if fullname.startswith("chalk.features"):
            return [(10, f"chalk.feature_n.feature_{str(n)}.feature", -1) for n in range(1, 257)]
        return []


def plugin(version: str):
    return ChalkPlugin



from datetime import timedelta
from typing import Any, Callable, Generic, List, Literal, Mapping, Optional, Set, Type, TypeVar, Union

import polars as pl

from chalk._validation.feature_validation import FeatureValidation
from chalk.utils.collections import ensure_tuple
from chalk.utils.duration import Duration, parse_chalk_duration, timedelta_to_duration

T = TypeVar("T")


class WindowedInstance(Generic[T]):
    def __init__(self, values: Mapping[str, T]):
        self.values = values

    def __call__(self, period: str):
        return self.values[period]


class WindowedMeta(type, Generic[T]):
    _kind: Type[T]

    def __getitem__(cls, underlying: Type[T]) -> "Windowed[Type[T]]":
        cls._kind = underlying
        return Windowed(
            buckets=[],
            mode="tumbling",
            description=None,
            owner=None,
            tags=None,
            name=None,
            default=None,
            max_staleness=None,
            version=None,
            etl_offline_to_online=None,
            encoder=None,
            decoder=None,
            min=None,
            max=None,
            min_length=None,
            max_length=None,
            contains=None,
            dtype=None,
        )  # noqa


JsonValue = Any


def get_duration_secs(duration: Union[str, int, timedelta]) -> int:
    if isinstance(duration, str):
        duration = parse_chalk_duration(duration)
    if isinstance(duration, timedelta):
        duration_secs_float = duration.total_seconds()
        duration_secs_int = int(duration_secs_float)
        if duration_secs_float != duration_secs_int:
            raise ValueError("Windows that are fractions of seconds are not yet supported")
        duration = duration_secs_int
    return duration


def get_name_with_duration(name_or_fqn: str, duration: Union[str, int, timedelta]) -> str:
    duration_secs = get_duration_secs(duration)
    return f"{name_or_fqn}__{duration_secs}__"


class Windowed(Generic[T], metaclass=WindowedMeta):
    _kind: Type[T]
    _buckets: List[str]
    _mode: Optional[str]
    _description: Optional[str]
    _owner: Optional[str]
    _tags: Optional[Any]
    _name: Optional[str]
    _default: Optional[T]
    _version: Optional[int]
    _etl_offline_to_online: Optional[bool]
    _encoder: Optional[Callable[[T], JsonValue]]
    _decoder: Optional[Callable[[JsonValue], T]]
    _min: Optional[T]
    _max: Optional[T]
    _min_length: Optional[int]
    _max_length: Optional[int]
    _contains: Optional[T]
    _dtype: Optional[Union[Type[pl.DataType], pl.DataType]]

    @property
    def buckets_seconds(self) -> Set[int]:
        return set(int(parse_chalk_duration(bucket).total_seconds()) for bucket in self._buckets)

    def to_feature(
        self,
        bucket: Optional[Union[int, str]],
    ):
        from chalk.features import Feature

        assert self._name is not None

        if bucket is None:
            name = self._name
        else:
            if get_duration_secs(bucket) not in self.buckets_seconds:
                raise ValueError(f"Bucket {bucket} is not in the list of specified buckets")
            name = get_name_with_duration(self._name, bucket)

        return Feature(
            name=name,
            version=self._version,
            owner=self._owner,
            tags=None if self._tags is None else list(ensure_tuple(self._tags)),
            description=self._description,
            primary=False,
            default=self._default,
            max_staleness=(
                timedelta_to_duration(self._max_staleness)
                if isinstance(self._max_staleness, timedelta)
                else self._max_staleness
            ),
            etl_offline_to_online=self._etl_offline_to_online,
            encoder=self._encoder,
            decoder=self._decoder,
            polars_dtype=self._dtype,
            validations=FeatureValidation(
                min=self._min,
                max=self._max,
                min_length=self._min_length,
                max_length=self._max_length,
                contains=self._contains,
            ),
            # Only the root feature should have all the durations
            # The pseudofeatures, which are bound to a duration, should not have the durations
            # of the other buckets
            window_durations=tuple(self.buckets_seconds) if bucket is None else tuple(),
            window_duration=None if bucket is None else get_duration_secs(bucket),
            window_mode=self._mode,
        )

    def __init__(
        self,
        buckets: List[str],
        mode: Union[Literal["tumbling"], Literal["continuous"]],
        description: Optional[str],
        owner: Optional[str],
        tags: Optional[Any],
        name: Optional[str],
        default: Optional[T],
        max_staleness: Optional[Duration],
        version: Optional[int],
        etl_offline_to_online: Optional[bool],
        encoder: Optional[Callable[[T], JsonValue]],
        decoder: Optional[Callable[[JsonValue], T]],
        min: Optional[T],
        max: Optional[T],
        min_length: Optional[int],
        max_length: Optional[int],
        contains: Optional[T],
        dtype: Optional[Union[Type[pl.DataType], pl.DataType]],
    ):
        self._name: Optional[str] = None
        self._buckets = buckets
        self._mode = mode
        self._description = description
        self._owner = owner
        self._tags = tags
        self._name = name
        self._default = default
        self._max_staleness = max_staleness
        self._description = description
        self._version = version
        self._etl_offline_to_online = etl_offline_to_online
        self._encoder = encoder
        self._decoder = decoder
        self._min: Optional[T] = min
        self._max: Optional[T] = max
        self._min_length: Optional[int] = min_length
        self._max_length: Optional[int] = max_length
        self._contains: Optional[T] = contains
        self._dtype: Optional[Union[Type[pl.DataType], pl.DataType]] = dtype


class SelectedWindow:
    def __init__(self, kind: Windowed, selected: str):
        self.windowed = kind
        self.selected = selected


def windowed(
    *buckets: str,
    mode: Union[Literal["tumbling"], Literal["continuous"]] = "tumbling",
    description: Optional[str] = None,
    owner: Optional[str] = None,
    tags: Optional[Any] = None,
    name: Optional[str] = None,
    default: Optional[T] = None,
    max_staleness: Optional[Duration] = ...,
    version: Optional[int] = None,
    etl_offline_to_online: Optional[bool] = None,
    encoder: Optional[Callable[[T], JsonValue]] = None,
    decoder: Optional[Callable[[JsonValue], T]] = None,
    min: Optional[T] = None,
    max: Optional[T] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    contains: Optional[T] = None,
    dtype: Optional[Union[Type[pl.DataType], pl.DataType]] = None,
) -> Windowed[T]:
    return Windowed(
        list(buckets),
        mode=mode,
        description=description,
        owner=owner,
        tags=tags,
        name=name,
        default=default,
        max_staleness=max_staleness,
        version=version,
        etl_offline_to_online=etl_offline_to_online,
        encoder=encoder,
        decoder=decoder,
        min=min,
        max=max,
        min_length=min_length,
        max_length=max_length,
        contains=contains,
        dtype=dtype,
    )



from __future__ import annotations

import copy
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union

from chalk.features.filter import Filter
from chalk.serialization.parsed_annotation import ParsedAnnotation
from chalk.streams import get_name_with_duration
from chalk.utils.collections import ensure_tuple

if TYPE_CHECKING:
    from chalk.features.feature_field import Feature

T = TypeVar("T")

__all__ = ["FeatureWrapper", "unwrap_feature"]


class FeatureWrapper:
    """
    FeatureWrapper emulates DataFrames and
    nested has-one relationships when used
    as a type annotation or within a filter.
    """

    def __init__(self, feature: Feature) -> None:
        # Binding as a private variable as not to have naming conflicts user's use of __getattr__
        self._chalk_feature = feature

    def __hash__(self):
        return hash(self._chalk_feature)

    def __gt__(self, other: object):
        return Filter(self._chalk_feature, ">", other)

    def __ge__(self, other: object):
        return Filter(self._chalk_feature, ">=", other)

    def __lt__(self, other: object):
        return Filter(self._chalk_feature, "<", other)

    def __le__(self, other: object):
        return Filter(self._chalk_feature, "<=", other)

    def _cmp(self, op: str, other: object):
        from chalk.features.feature_field import Feature

        if isinstance(other, Feature):
            # If comparing against a feature directly, then we know it's not being used in a join condition
            # Since join conditions would be against another FeatureWrapper or a literal value
            is_eq = self._chalk_feature == other
            # They are the same feature. Short-circuit and return a boolean
            if op == "==" and is_eq:
                return True
            if op == "!=" and not is_eq:
                return False
            return NotImplemented  # GT / LT doesn't really make sense otherwise
        return Filter(self._chalk_feature, op, other)

    def __ne__(self, other: object):
        return self._cmp("!=", other)

    def __eq__(self, other: object):
        return self._cmp("==", other)

    def __and__(self, other: object):
        return self._cmp("and", other)

    def __or__(self, other: object):
        if other is None:
            return Optional[self]
        return self._cmp("or", other)

    def __repr__(self):
        return f"FeatureWrapper(fqn={self._chalk_feature.namespace}.{self._chalk_feature.name}, typ={self._chalk_feature.typ})"

    def __str__(self):
        return str(self._chalk_feature)

    def in_(self, examples: Iterable):
        return self._cmp("in", examples)

    def __call__(self, *args: Any, **kwargs: Any):
        # Using a generic signature since this signature must support all types of features
        # Currently, though, only windowed features are callable
        if self._chalk_feature.is_windowed:
            return self._chalk_get_windowed_feature(*args, **kwargs)
        raise TypeError(f"Feature {self} is not callable")

    def _chalk_get_windowed_feature(self, window: Union[str, int]):
        if not isinstance(window, (str, int)):
            raise TypeError("Window duration must be a string or an int")

        from chalk.features import FeatureSetBase

        parent = (
            FeatureSetBase.registry[self._chalk_feature.namespace]
            if len(self._chalk_feature.path) == 0
            else FeatureWrapper(self._chalk_feature.path[-1].parent)
        )
        desired_attribute_name = get_name_with_duration(self._chalk_feature.attribute_name, window)
        if not hasattr(parent, desired_attribute_name):
            formatted_window_durations = [f"'{x}s'" for x in self._chalk_feature.window_durations]
            raise TypeError(
                (
                    f"Unsupported window duration '{window}' for '{self._chalk_feature.root_fqn}'. "
                    f"Durations {', '.join(formatted_window_durations)} are supported."
                )
            )
        return getattr(parent, desired_attribute_name)

    def __getitem__(self, item: Any):
        from chalk.features.dataframe import DataFrame

        if len(self._chalk_feature.window_durations) > 0:
            return self._chalk_get_windowed_feature(*ensure_tuple(item))

        if self._chalk_feature.typ is not None and issubclass(self._chalk_feature.typ.parsed_annotation, DataFrame):
            f_copy = FeatureWrapper(copy.copy(self._chalk_feature))

            f_copy._chalk_feature.typ = ParsedAnnotation(underlying=self._chalk_feature.typ.parsed_annotation[item])

            return f_copy

        raise TypeError(f"Feature {self} does not support subscripting")

    def __getattr__(self, item: str):
        from chalk.features.feature_field import Feature
        from chalk.features.feature_set import Features

        # Passing through __getattr__ on has_one features, as users can use getattr
        # notation in annotations for resolvers
        if item.startswith("__"):
            # Short-circuiting on the dunders to be compatible with copy.copy
            raise AttributeError(item)

        if self._chalk_feature.typ is not None and issubclass(self._chalk_feature.typ.underlying, Features):
            for f in self._chalk_feature.typ.underlying.features:
                assert isinstance(f, Feature), f"HasOne feature {f} does not inherit from FeaturesBase"
                if f.attribute_name == item:
                    return FeatureWrapper(self._chalk_feature.copy_with_path(f))
        raise AttributeError(f"'{self}' has no attribute '{item}'")


def unwrap_feature(maybe_feature_wrapper: Any) -> Feature:
    """Unwrap a class-annotated FeatureWrapper instance into the underlying feature.

    For example:

    .. code-block::

        @features
        class FooBar:
            foo: str
            bar: int

        type(FooBar.foo) is FeatureWrapper
        type(unwrap_feature(FooBar.foo)) is Feature
    """
    from chalk.features.feature_field import Feature

    if isinstance(maybe_feature_wrapper, FeatureWrapper):
        maybe_feature_wrapper = maybe_feature_wrapper._chalk_feature
    if isinstance(maybe_feature_wrapper, Feature):
        return maybe_feature_wrapper
    raise TypeError(
        f"{maybe_feature_wrapper} is of type {type(maybe_feature_wrapper).__name__}, expecting type FeatureWrapper"
    )



from datetime import datetime

from chalk.client.client_impl import _expand_scalar_features_shallow
from chalk.features import DataFrame, feature_time, features, has_many, has_one


@features
class Ankle:
    id: int
    foot_id: int


@features
class Foot:
    id: int
    is_flat: bool
    person_id: int
    ts: datetime = feature_time()
    ankle: Ankle = has_one(lambda: Ankle.foot_id == Foot.id)


@features
class Person:
    id: int
    name: str
    best_foot_id: int
    best_foot: Foot = has_one(lambda: Foot.id == Person.best_foot_id)
    feet: DataFrame[Foot] = has_many(lambda: Person.id == Foot.person_id)


def _expand(*items):
    return set(_expand_scalar_features_shallow(items))


def test_expand_features():
    assert _expand(Person) == {"person.id", "person.name", "person.best_foot_id"}
    assert _expand(Ankle) == {"ankle.id", "ankle.foot_id"}
    assert _expand(Foot) == {"foot.id", "foot.is_flat", "foot.person_id"}


def test_expand_has_ones():
    assert _expand(Foot.ankle) == {"foot.ankle.id", "foot.ankle.foot_id"}
    assert _expand(Person.best_foot.ankle) == {"person.best_foot.ankle.id", "person.best_foot.ankle.foot_id"}
    assert _expand(Foot.ankle, Foot) == {
        "foot.ankle.id",
        "foot.ankle.foot_id",
        "foot.id",
        "foot.is_flat",
        "foot.person_id",
    }
    assert _expand(Person, Person.best_foot) == {
        "person.id",
        "person.name",
        "person.best_foot_id",
        "person.best_foot.id",
        "person.best_foot.is_flat",
        "person.best_foot.person_id",
    }
    assert _expand(Person, Person.best_foot, Person.best_foot.ankle) == {
        "person.id",
        "person.name",
        "person.best_foot_id",
        "person.best_foot.id",
        "person.best_foot.is_flat",
        "person.best_foot.person_id",
        "person.best_foot.ankle.id",
        "person.best_foot.ankle.foot_id",
    }



from __future__ import annotations

import inspect

import collections.abc
import dataclasses
import datetime
import enum
import functools
import inspect
import operator
import pathlib
import warnings
from pydantic import BaseModel
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import polars as pl
from pydantic import BaseModel

try:
    import zoneinfo
except ImportError:
    # Zoneinfo was introduced in python 3.9
    from backports import zoneinfo

from dateutil import parser

from chalk.features.feature_field import Feature
from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
from chalk.features.filter import Filter, TimeDelta
from chalk.serialization.codec import FEATURE_CODEC
from chalk.utils.collections import ensure_tuple, get_unique_item

if TYPE_CHECKING:
    import pandas as pd
    import pyarrow

    from chalk.features.feature_set import Features

else:
    try:
        import pandas as pd
    except ImportError:
        pd = None
    try:
        import pyarrow
    except ImportError:
        pyarrow = None

TDataFrame = TypeVar("TDataFrame", bound="DataFrame")


def _iso_parse(x: str, expected_tz: datetime.tzinfo | None) -> datetime.datetime:
    """This function converts the input into an expected timezone, and drops it, so we can cast it in polars with with_time_zone"""
    parsed = parser.isoparse(x)
    if parsed.tzinfo is not None and expected_tz is not None:
        # Convert the timezone
        parsed = parsed.astimezone(expected_tz)
    # Now drop the timezone
    return parsed.replace(tzinfo=None)


class DataFrameMeta(type):
    def __getitem__(cls, item: Any) -> Type[DataFrame]:
        from chalk.features.feature_set import Features

        # leaving untyped as we type the individual features as their object type
        # but item should really be Filter (expressions), Feature classes, or Feature instances
        cls = cast(Type[DataFrame], cls)

        item = ensure_tuple(item)

        # Disallow string annotations like DataFrame["User"].
        # Instead, the entire thing should be in quotes -- like "DataFrame[User]"
        for x in item:
            if isinstance(x, str):
                raise TypeError(
                    (
                        f'Annotation {cls.__name__}["{x}", ...] is unsupported. Instead, use a string for the entire annotation -- for example: '
                        f'"{cls.__name__}[{x}, ...]"'
                    )
                )

        # If doing multiple subscript, then keep the filters, but do not keep the individual columns
        # TODO: Validate that any new columns are a subset of the existing columns
        item = [*item, *cls.filters]

        new_filters: List[Filter] = []
        new_references_feature_set: Optional[Type[Features]] = None
        new_columns: List[Feature] = []
        pydantic_model = None

        for a in item:
            if isinstance(a, Filter):
                new_filters.append(a)
            elif isinstance(a, type) and issubclass(a, Features):
                if new_references_feature_set is not None:
                    raise ValueError(
                        f"Multiple referenced feature sets -- {new_references_feature_set} and {a} -- are not supported."
                    )
                new_references_feature_set = a
            elif isinstance(a, Feature):
                new_columns.append(a)
            elif isinstance(a, FeatureWrapper):
                new_columns.append(a._chalk_feature)
            elif isinstance(a, bool):
                # If we encounter a bool, that means we are evaluating the type annotation before
                # the ResolverAstParser had a chance to extract the source and rewrite the and/or/in operations
                # into expressions that return filters instead of booleans
                # This function will be called again for this annotation, so we can ignore it for now.
                pass
            elif inspect.isclass(a) and issubclass(a, BaseModel):
                pydantic_model = a
            else:
                raise TypeError(f"Invalid type for DataFrame[{a}]: {type(a)}")

        if len(new_columns) == 0 and new_references_feature_set is None:
            # This is possible if you have something like
            # Users.transactions[after('60d')]
            # In this case, keep all existing columns
            # But if you did
            # Users.transactions[Transaction.id, after('60d')]
            # Then keep only the id column
            new_columns = list(cls.__columns__)
            new_references_feature_set = cls.__references_feature_set__

        class SubclassedDataFrame(cls):
            filters = tuple(new_filters)
            __columns__ = tuple(new_columns)
            __references_feature_set__ = new_references_feature_set
            __pydantic_model__ = pydantic_model

            def __new__(cls: Type[TDataFrame], *args: Any, **kwargs: Any) -> TDataFrame:
                raise RuntimeError(
                    "A SubclassedDataFrame should never be instantiated. Instead, instantiate a DataFrame(...)."
                )

        return SubclassedDataFrame

    def __repr__(cls):
        cls = cast(Type[DataFrame], cls)
        elements = [str(x) for x in (*cls.filters, *cls.columns)]
        return f"DataFrame[{', '.join(elements)}]"

    @property
    def columns(cls) -> Tuple[Feature, ...]:
        # Computing the columns lazily as we need to implicitly parse the type annotation
        # to determine if a field is a has-many, and we don't want to do that on the
        # __getitem__ which could happen before forward references can be resolved
        # So, using a property on the metaclass, which acts like an attribute on the class, to
        # provide the dataframe columns
        from chalk.features.feature_field import Feature

        cls = cast(Type[DataFrame], cls)
        columns: Set[Feature] = set()
        for x in cls.__columns__:
            assert isinstance(x, Feature)
            # If a feature is directly specified, allow has-ones. But still do not allow has-many features
            assert not x.is_has_many, "Has-many features are not allowed to be specified within a DataFrame"
            columns.add(x)
        if cls.__references_feature_set__ is not None:
            # Only include the first-level feature types
            # Do not recurse has-ones and has-many as that could create an infinite loop
            for x in cls.__references_feature_set__.features:
                assert isinstance(x, Feature)
                if not x.is_has_many and not x.is_has_one:
                    columns.add(x)
        return tuple(columns)

    @property
    def references_feature_set(cls):
        from chalk.features.feature_set import FeatureSetBase

        cls = cast(Type[DataFrame], cls)
        if cls.__references_feature_set__ is not None:
            return cls.__references_feature_set__
        else:
            # Determine the unique @features cls that encompasses all columns
            root_ns = get_unique_item((x.root_namespace for x in cls.__columns__), "root ns")
        return FeatureSetBase.registry[root_ns]

    @property
    def namespace(cls) -> str:
        cls = cast(Type[DataFrame], cls)
        namespaces = [x.path[0].parent.namespace if len(x.path) > 0 else x.namespace for x in cls.columns]
        # Remove the pseudo-columns
        namespaces = [x for x in namespaces if not x.startswith("__chalk__")]
        return get_unique_item(namespaces, f"dataframe {cls.__name__} column namespaces")


class DataFrame(metaclass=DataFrameMeta):
    filters: ClassVar[Tuple[Filter, ...]] = ()
    columns: Tuple[Feature, ...]  # set via a @property on the metaclass
    __columns__: ClassVar[Tuple[Feature, ...]] = ()
    references_feature_set: Optional[Type[Features]]  # set via a @property on the metaclass
    __references_feature_set__: ClassVar[Optional[Type[Features]]] = None
    __pydantic_model__: ClassVar[Optional[Type[BaseModel]]] = None

    def __init__(
        self,
        data: Union[
            Dict[Union[str, Feature, FeatureWrapper, Any], Sequence[Features]],
            Sequence[Any],
            pl.DataFrame,
            pl.LazyFrame,
            pd.DataFrame,
            Any,  # Polars supports a bunch of other formats for initialization of a DataFrame
        ] = None,
        pandas_dataframe: Optional[pd.DataFrame] = None,  # For backwards compatibility
        convert_dtypes: bool = True,  # By default, data should match the dtype of the feature. However, when doing comparisions, data will be converted to bools, in which case it should no longer be validated.
        pydantic_model: Optional[Type[BaseModel]] = None,
    ):
        """Construct a Chalk DataFrame

        Args:
            data: The data. Can be an existing Pandas DataFrame, Polars DataFrame or LazyFrame, a sequence of feature instances,
                or a dict mapping a feature to a sequence of values.
            convert_dtypes: Whether to convert the data to match the dtype of the underlying feature.

        """
        from chalk.features.feature_set import Features, FeatureSetBase

        self._pydantic_model = pydantic_model

        # Typing the keys of ``data`` as Any, as {FeatureCls.item: x} would be typed as the underlying annotation of the features cls
        if pandas_dataframe is not None:
            warnings.warn(
                DeprecationWarning("ChalkDataFrameImpl(pandas_dataframe=...) has been renamed to DataFrame(data=...)")
            )
            data = pandas_dataframe
        if pd is not None and isinstance(data, pd.DataFrame):
            # Convert the columns to root fqn strings
            # str(Feature) and str(FeatureWrapper) return the root fqns
            data = data.rename(columns={k: str(k) for k in data.columns})
            assert isinstance(data, pd.DataFrame)
            data.columns = data.columns.astype("string")
            data = pl.from_pandas(data)
        if not isinstance(data, (pl.LazyFrame, pl.DataFrame)):
            if isinstance(data, (collections.abc.Sequence)) and not isinstance(data, str):
                # If it is a sequence, it could be a sequence of feature classes instances
                # If so, set the columns by inspecting the feature classes
                # If columns are none, then inspect the data to determine if they are feature classes
                # Otherwise, if the columns are specified, do not introspect the list construction
                features_typ = None
                new_data: dict[str, list[Any]] = {}
                for row in data:
                    if not isinstance(row, Features):
                        raise ValueError("If specifying data as a sequence, it must be a sequence of Features")
                    if features_typ is None:
                        features_typ = type(row)
                        for x in row.features:
                            assert isinstance(x, Feature)
                            assert x.attribute_name is not None
                            try:
                                feature_val = getattr(row, x.attribute_name)
                            except AttributeError:
                                continue
                            new_data[x.root_fqn] = []

                    if features_typ != type(row):
                        raise ValueError("Cannot mix different feature classes in a DataFrame")
                    for x in row.features:
                        assert isinstance(x, Feature)
                        assert x.attribute_name is not None
                        try:
                            feature_val = getattr(row, x.attribute_name)
                        except AttributeError:
                            if x.root_fqn in new_data:
                                raise ValueError(f"Feature {x.root_fqn} is not defined in all feature sets.")
                            continue
                        if x.is_has_many:
                            raise ValueError("DataFrames within DataFrames are not supported")
                        if x.root_fqn not in new_data:
                            raise ValueError(f"Feature {x.root_fqn} is not defined in all feature sets.")
                        new_data[x.root_fqn].append(feature_val)
                data = new_data
            if isinstance(data, dict):
                # Convert the columns to root fqn strings
                # str(Feature) and str(FeatureWrapper) return the root fqns
                new_data_dict: Dict[str, Sequence[Any] | pl.Series] = {}
                for (k, v) in data.items():
                    v = [_preprocess_element(x) for x in v]

                    if convert_dtypes:
                        polars_dtype = FEATURE_CODEC.get_polars_dtype(str(k))
                        elements_as_series = _generate_empty_series_for_dtype(str(k), polars_dtype, 0)
                        null_element = _generate_empty_series_for_dtype(str(k), polars_dtype, 1)
                        for i, element in enumerate(v):
                            if element is None or isinstance(element, (list, tuple)) and len(element) == 0:
                                elements_as_series.append(null_element)
                            else:
                                element_as_series = pl.Series(name=str(k), values=[element])
                                if element_as_series.dtype != polars_dtype:

                                    if _polars_dtype_contains_struct(polars_dtype):
                                        # Cannot cast to a struct type. Instead, will error, so the user can ensure the underlying
                                        # dictionaries / dataclasses are of the correct type
                                        raise TypeError(
                                            (
                                                f"Expected field `{str(k)}` to have dtype `{str(polars_dtype)}`; got dtype `{str(element_as_series.dtype)}`"
                                                f" for element {element} at index {i}"
                                            )
                                        )
                                    else:
                                        try:
                                            if (
                                                isinstance(element_as_series.dtype, type)
                                                and issubclass(element_as_series.dtype, pl.Utf8)
                                                and isinstance(polars_dtype, pl.Datetime)
                                            ):
                                                if polars_dtype.tz is None:
                                                    element_as_series = element_as_series.apply(
                                                        lambda x: None if x is None else _iso_parse(x, None),
                                                    ).dt.with_time_zone(None)
                                                else:
                                                    tzinfo = zoneinfo.ZoneInfo(polars_dtype.tz)
                                                    element_as_series = element_as_series.apply(
                                                        lambda x: None if x is None else _iso_parse(x, tzinfo),
                                                    ).dt.with_time_zone(polars_dtype.tz)
                                            else:
                                                element_as_series = element_as_series.cast(polars_dtype)
                                        except (pl.ComputeError, pl.InvalidOperationError) as e:
                                            # If operating on a lazy frame, we won't get the exception until we call .compute()
                                            raise TypeError(
                                                f"Feature `{str(k)}` at index {i} with value '{element}' could not be converted to dtype `{polars_dtype.string_repr()}`"
                                            ) from e
                                elements_as_series.append(element_as_series)
                        elements_as_series = elements_as_series.rechunk(True)
                    else:
                        elements_as_series = pl.Series(name=str(k), values=v)
                    # # Otherwise, we can let polars infer the dtype. We will cast it later if convert_dtypes is True
                    # # We do not want to specify it now, as otherwise polars will drop values instead of cast
                    new_data_dict[str(k)] = elements_as_series
                data = new_data_dict
            data = pl.DataFrame(data)
        if isinstance(data, (pl.LazyFrame, pl.DataFrame)):
            underlying = data
        else:
            raise ValueError(f"Unable to convert data of type {type(data).__name__} into a DataFrame")
        # Rename / validate that all column names are root fqns
        # It is possible that a feature name is a
        if self._pydantic_model is None:
            self.columns = tuple(Feature.from_root_fqn(str(c)) for c in underlying.columns)
            underlying = underlying.rename(
                {original_c: new_c.root_fqn for (original_c, new_c) in zip(underlying.columns, self.columns)}
            )
        else:
            self.columns = ()

        # Convert columns to the correct dtype to match the fqn
        if self._pydantic_model is None and convert_dtypes:
            for root_fqn in underlying.columns:
                expected_dtype = FEATURE_CODEC.get_polars_dtype(root_fqn)
                actual_dtype = underlying.schema[root_fqn]
                if actual_dtype != expected_dtype:
                    if _polars_dtype_contains_struct(expected_dtype):
                        # Cannot cast to a struct type. Instead, will error, so the user can ensure the underlying
                        # dictionaries / dataclasses are of the correct type
                        raise TypeError(
                            f"Expected field `{root_fqn}` to have dtype `{expected_dtype}`; got dtype `{actual_dtype}`"
                        )
                    else:
                        try:
                            if isinstance(expected_dtype, pl.Datetime) and actual_dtype == pl.Utf8:
                                tzinfo = None if expected_dtype.tz is None else zoneinfo.ZoneInfo(expected_dtype.tz)
                                if expected_dtype.tz is None:
                                    col = (
                                        pl.col(root_fqn)
                                        .apply(
                                            lambda x: None if x is None else _iso_parse(cast(str, x), None),
                                        )
                                        .dt.with_time_zone(None)
                                    )
                                else:
                                    tzinfo = zoneinfo.ZoneInfo(expected_dtype.tz)
                                    col = (
                                        pl.col(root_fqn)
                                        .apply(
                                            lambda x: None if x is None else _iso_parse(cast(str, x), tzinfo),
                                        )
                                        .dt.with_time_zone(expected_dtype.tz)
                                    )
                                underlying = underlying.with_columns([col])
                            else:
                                underlying = underlying.with_columns([pl.col(root_fqn).cast(expected_dtype)])
                        except pl.ComputeError as e:
                            # If operating on a lazy frame, we won't get the exception until we call .compute()
                            raise TypeError(
                                f"Values for feature `{root_fqn}` could not be converted to dtype `{expected_dtype.string_repr()}`. Found type {actual_dtype}, instead."
                            ) from e

        if isinstance(underlying, pl.DataFrame):
            underlying = underlying.lazy()
        self._underlying: pl.LazyFrame = underlying

        # Remove the pseudo-features when determining the namespace
        namespaces = (x.path[0].parent.namespace if len(x.path) > 0 else x.namespace for x in self.columns)
        namespaces_set = set(x for x in namespaces if not x.startswith("__chalk__"))
        if len(namespaces_set) == 1:
            self.namespace = get_unique_item(namespaces_set, f"dataframe column namespaces")
            self.references_feature_set = FeatureSetBase.registry[self.namespace]
        else:
            # Allow empty dataframes or dataframes with multiple namespaces
            self.namespace = None
            self.references_feature_set = None

    ##############
    # Classmethods
    ##############

    @classmethod
    def from_dict(
        cls: Type[TDataFrame],
        data: Dict[Union[str, Feature, FeatureWrapper, Any], Sequence[Any]],
    ) -> TDataFrame:
        warnings.warn(DeprecationWarning("DataFrame.from_dict(...) is deprecated. Instead, use DataFrame(...)"))
        df = cls(data)
        return df

    @overload
    @classmethod
    def from_list(
        cls: Type[TDataFrame],
        data: Sequence[Features],
        /,
    ) -> TDataFrame:
        ...

    @overload
    @classmethod
    def from_list(cls: Type[TDataFrame], *data: Features) -> TDataFrame:
        ...

    @classmethod
    def from_list(cls: Type[TDataFrame], *data: Union[Features, Sequence[Features]]) -> TDataFrame:
        warnings.warn(DeprecationWarning("DataFrame.from_list(...) is deprecated. Instead, use DataFrame(...)"))
        if len(data) == 1 and isinstance(data[0], collections.abc.Sequence):
            # Passed a list as the first argument
            features_seq = data[0]
        else:
            data = cast("Tuple[Features]", data)
            features_seq = data
        df = cls(features_seq)
        return df

    @classmethod
    def read_parquet(
        cls: Type[TDataFrame],
        path: Union[str, pathlib.Path],
        columns: Optional[Union[List[int], List[str], Dict[str, Union[str, Feature, FeatureWrapper, Any]]]] = None,
    ) -> TDataFrame:
        if isinstance(columns, dict):
            columns_to_read = list(columns.keys())
            column_map = {k: str(v) for (k, v) in columns.items()}
        else:
            columns_to_read = columns
            column_map = None
        data = pl.read_parquet(path, columns_to_read)
        if column_map is not None:
            data = data.rename(column_map)
        return cls(data)

    @classmethod
    def read_csv(
        cls: Type[TDataFrame],
        path: Union[str, pathlib.Path],
        has_header: bool,
        columns: Optional[Union[List[int], List[str], Dict[str, Union[str, Feature, FeatureWrapper, Any]]]] = None,
    ) -> TDataFrame:
        if isinstance(columns, dict):
            columns_to_read = list(columns.keys())
            column_map = {k: str(v) for (k, v) in columns.items()}
        else:
            columns_to_read = columns
            column_map = None
        data = pl.read_csv(path, has_header, columns_to_read)
        if column_map is not None:
            data = data.rename(column_map)
        return cls(data)

    #############
    # Aggregation
    #############

    def max(self):
        return DataFrame(self._underlying.max(), convert_dtypes=False, pydantic_model=self._pydantic_model)

    def mean(self):
        return DataFrame(self._underlying.mean(), convert_dtypes=False, pydantic_model=self._pydantic_model)

    def median(self):
        return DataFrame(self._underlying.median(), convert_dtypes=False, pydantic_model=self._pydantic_model)

    def min(self):
        return DataFrame(self._underlying.min(), convert_dtypes=False, pydantic_model=self._pydantic_model)

    def std(self, ddof: int = 1):
        return DataFrame(self._underlying.std(ddof), convert_dtypes=False, pydantic_model=self._pydantic_model)

    def sum(self):
        # Treat missing sums as zero
        return DataFrame(self._underlying.sum().fill_null(0), convert_dtypes=False, pydantic_model=self._pydantic_model)

    def var(self, ddof: int = 1):
        return DataFrame(self._underlying.var(ddof), convert_dtypes=False, pydantic_model=self._pydantic_model)

    ####################
    # Summary Operations
    ####################

    # These ops require us to materialize the dataframe.

    def _materialize(self) -> pl.DataFrame:
        materialized = self._underlying.collect()
        self._underlying = materialized.lazy()
        return materialized

    def any(self):
        """Returns whether any of the values in the dataframe are truthy. Requires the dataframe to only contain boolean values."""
        if not all(isinstance(x, type) and issubclass(x, pl.Boolean) for x in self._underlying.dtypes):
            raise TypeError("DataFrame.any() is not defined on a dataframe that contains non-boolean columns.")
        materialized = self._materialize()
        return any(col.any() for col in materialized.get_columns())

    def all(self):
        """Returns whether all of the values in the dataframe are truthy. Requires the dataframe to only contain boolean values."""
        if not all(isinstance(x, type) and issubclass(x, pl.Boolean) for x in self._underlying.dtypes):
            raise TypeError("DataFrame.any() is not defined on a dataframe that contains non-boolean columns.")
        materialized = self._materialize()
        return all(col.all() for col in materialized.get_columns())

    def __len__(self):
        materialized = self._materialize()
        return len(materialized)

    @property
    def shape(self):
        materialized = self._materialize()
        return materialized.shape

    def item(self):
        """Get the only item from the dataframe."""
        materialized = self._materialize()
        if materialized.shape == (1, 1):
            return materialized.rows()[0][0]
        raise ValueError(
            "The dataframe contains multiple items. DataFrame.item() can only be used if the dataframe has a single element."
        )

    def __bool__(self):
        if self.shape == (1, 1):
            # It's a dataframe of 1 item. self.any() and self.all() would return the same thing
            return self.all()
        raise ValueError("__bool__ is ambiguous on a DataFrame. Instead, use DataFrame.any() or DataFrame.all().")

    def __str__(self):
        materialized = self._materialize()
        return str(materialized)

    def __repr__(self):
        materialized = self._materialize()
        return repr(materialized)

    def __float__(self):
        return float(self.item())

    def __int__(self):
        return int(self.item())

    ############################
    # Arithmetic and Comparisons
    ############################

    # These ops require us to materialize the dataframe.

    def _perform_op(
        self,
        op: Callable[[Any, Any], Any],
        other: Union[DataFrame, pl.DataFrame, pd.DataFrame, Any],
        convert_dtypes: bool,
    ):
        materialized = self._materialize()
        if isinstance(other, DataFrame):
            other = other.to_polars()
        if isinstance(other, pl.LazyFrame):
            other = other.collect()
        if isinstance(other, pd.DataFrame):
            other = pl.from_pandas(other)
        return DataFrame(
            op(materialized, other),
            convert_dtypes=convert_dtypes,
            pydantic_model=self._pydantic_model,
        )

    def __eq__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):  # type: ignore
        return self._perform_op(operator.eq, other, convert_dtypes=False)

    def __ne__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):  # type: ignore
        return self._perform_op(operator.ne, other, convert_dtypes=False)

    def __gt__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):
        return self._perform_op(operator.gt, other, convert_dtypes=False)

    def __lt__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):
        return self._perform_op(operator.lt, other, convert_dtypes=False)

    def __ge__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):
        return self._perform_op(operator.ge, other, convert_dtypes=False)

    def __le__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):
        return self._perform_op(operator.le, other, convert_dtypes=False)

    def __add__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]) -> DataFrame:
        return self._perform_op(operator.add, other, convert_dtypes=True)

    def __sub__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]) -> DataFrame:
        return self._perform_op(operator.sub, other, convert_dtypes=True)

    def __mul__(self, other: Union[int, float]) -> DataFrame:
        return self._perform_op(operator.mul, other, convert_dtypes=True)

    def __truediv__(self, other: Union[int, float]) -> DataFrame:
        return self._perform_op(operator.truediv, other, convert_dtypes=True)

    def __floordiv__(self, other: Union[int, float]) -> DataFrame:
        return self._perform_op(operator.floordiv, other, convert_dtypes=True)

    def __mod__(self, other: Union[int, float]) -> DataFrame:
        return self._perform_op(operator.mod, other, convert_dtypes=True)

    def __pow__(self, other: Union[int, float]) -> DataFrame:
        return self._perform_op(operator.pow, other, convert_dtypes=True)

    ############
    # Conversion
    ############
    def to_polars(self) -> pl.LazyFrame:
        """Get the underlying dataframe as a Polars LazyFrame."""
        # Implementing to_polars() to return the LazyFrame, rather than the DataFrame, to encourage customers to use
        # lazy operations if possible. They can manually call .collect() if needed
        return self._underlying

    def to_pandas(self) -> pd.DataFrame:
        """Get the underlying dataframe as a Pandas DataFrame."""
        # For pandas, the columns should be the Features, not the root fqns
        def types_mapper(dtype: pyarrow.DataType):
            if dtype in (pyarrow.utf8(), pyarrow.large_utf8()):
                return pd.StringDtype()
            return None

        pd_dataframe = self._underlying.collect().to_pandas(types_mapper=types_mapper)
        pd_dataframe.columns = pd_dataframe.columns.astype("object")
        pd_dataframe = pd_dataframe.rename(columns={x.root_fqn: x for x in self.columns})
        return pd_dataframe

    #######################
    # Filtering / Selecting
    #######################
    def __getitem__(self, item: Any):
        from chalk.features.feature_set import FeatureSetBase

        has_bool_or_filter_value = any(isinstance(x, (bool, Filter)) for x in ensure_tuple(item))
        if has_bool_or_filter_value:
            # If we have a boolean or Filter value, then that means we need to ast-parse the caller since
            # python has already evaluated AND, OR, and IN operations into literal booleans or Filters
            # Skipping the parsing unless if we have need to for efficiency and to eliminate conflicts
            # with pytest
            from chalk.df.ast_parser import parse_dataframe_getitem

            item = parse_dataframe_getitem()
        if any(isinstance(x, (FeatureWrapper, Feature, Filter)) for x in ensure_tuple(item)):
            # Use the Chalk projection / selection syntax, where we support our Filter objects and
            # selection by column name
            projections: list[str] = []
            filters: List[Filter] = []
            for x in ensure_tuple(item):
                if isinstance(x, (FeatureWrapper, Feature, str)):
                    projections.append(str(x))

                elif isinstance(x, Filter):
                    filters.append(x)
                else:
                    raise TypeError(
                        "When indexing by Filters or Features, it is not simultaneously possible to perform other indexing operations."
                    )

            now = datetime.datetime.now(tz=datetime.timezone.utc)
            timestamp_feature = None if self.namespace is None else FeatureSetBase.registry[self.namespace].__chalk_ts__
            pl_expr = convert_filters_to_pl_expr(filters, self._underlying.schema, timestamp_feature, now)
            df = self._underlying
            if pl_expr is not None:
                df = df.filter(pl_expr)
            # Do the projection
            if len(projections) > 0:
                df = df.select(projections)
            return DataFrame(df, pydantic_model=self)
        else:
            # Otherwise, use the standard polars selection format
            # Must materialize the dataframe to use __getitem__
            materialized = self._materialize()
            df = materialized[item]
            return DataFrame(df, pydantic_model=self._pydantic_model)


def _feature_type_or_value(e: Union[Feature, FeatureWrapper]):
    if isinstance(e, FeatureWrapper):
        e = unwrap_feature(e)
    return e


def _polars_dtype_contains_struct(dtype: pl.DataType | Type[pl.DataType]):
    """Returns whether the dtype contains a (potentially nested) struct"""
    if isinstance(dtype, pl.Struct) or (isinstance(dtype, type) and issubclass(dtype, pl.Struct)):
        return True
    if isinstance(dtype, pl.List):
        assert dtype.inner is not None
        return _polars_dtype_contains_struct(dtype.inner)
    return False


def _maybe_replace_timestamp_feature(f: Union[Feature, Any], observed_at_feature: Optional[Feature]):
    """Replace the ``CHALK_TS`` pseudo-feature with the actual timestamp column."""
    if not isinstance(f, Feature) or f.fqn != "__chalk__.CHALK_TS":
        return f
    if observed_at_feature is not None:
        return observed_at_feature
    raise ValueError("No Timestamp Feature Found")


def _maybe_convert_timedelta_to_timestamp(
    f: Union[TimeDelta, datetime.timedelta, Any], now: Optional[datetime.datetime]
):
    """Convert timedeltas relative to ``now`` into absolute datetimes."""
    if isinstance(f, TimeDelta):
        f = f.to_std()
    if isinstance(f, datetime.timedelta):
        if now is None:
            raise ValueError(
                "The filter contains a relative timestamp. The current datetime must be provided to evaluate this filter."
            )
        return now + f
    return f


def _parse_feature_or_value(
    f: Union[Feature, Any], timestamp_feature: Optional[Feature], now: Optional[datetime.datetime]
):
    """Parse a feature or value into the correct type that can be used for filtering."""
    f = _feature_type_or_value(f)
    f = _maybe_convert_timedelta_to_timestamp(f, now)
    f = _maybe_replace_timestamp_feature(f, timestamp_feature)
    if isinstance(f, enum.Enum):
        f = f.value
    return f


def _polars_is_eq(
    lhs: Any,
    rhs: Any,
    lhs_dtype: Optional[Union[Type[pl.DataType], pl.DataType]],
    rhs_dtype: Optional[Union[Type[pl.DataType], pl.DataType]],
):
    """Compare a column with another column or literal value, including possibly a struct.
    Polars does not permit equality comparisons on structs directly. Instead, must compare field by field, potentially recursively.
    """
    if rhs_dtype is not None:
        if lhs_dtype is None:
            # Swap the columns
            return _polars_is_eq(rhs, lhs, rhs_dtype, lhs_dtype)
        else:
            # Comparing two columns
            assert isinstance(lhs, pl.Expr)
            assert isinstance(rhs, pl.Expr)
            if isinstance(lhs_dtype, pl.Struct):
                assert isinstance(rhs_dtype, pl.Struct)
                # Assert equality field by field
                filters = []
                for field in lhs_dtype.fields:
                    field_name = field.name
                    new_lhs = lhs.struct.field(field_name)
                    new_lhs_dtype = field.dtype
                    new_rhs = rhs.struct.field(field_name)
                    new_rhs_dtype = field.dtype
                    filters.append(_polars_is_eq(new_lhs, new_rhs, new_lhs_dtype, new_rhs_dtype))
                assert len(filters) > 0, "structs with 0 fields are unsupported"
                return functools.reduce(lambda a, b: a & b, filters)
            return lhs == rhs
    # rhs is literal
    # lhs is a column
    assert lhs_dtype is not None, "one side must be a column"
    assert isinstance(lhs, pl.Expr)
    if isinstance(lhs_dtype, pl.Struct):
        # Assert equality field by field
        filters = []
        for field in lhs_dtype.fields:
            field_name = field.name
            new_lhs = lhs.struct.field(field_name)
            new_lhs_dtype = field.dtype
            # Assuming that struct-like objects make their members accessible by attribute name or __getitem__
            try:
                new_rhs = getattr(rhs, field_name)
            except AttributeError:
                new_rhs = rhs[field_name]
            new_rhs_dtype = None  # literal values do not have dtypes
            filters.append(_polars_is_eq(new_lhs, new_rhs, new_lhs_dtype, new_rhs_dtype))
        assert len(filters) > 0, "structs with 0 fields are unsupported"
        return functools.reduce(lambda a, b: a & b, filters)
    if not isinstance(lhs_dtype, type):
        lhs_dtype = type(lhs_dtype)
    return lhs == pl.lit(rhs, dtype=lhs_dtype, allow_object=True)


def _polars_is_in(lhs: pl.Expr, rhs: Iterable, lhs_dtype: Union[Type[pl.DataType], pl.DataType]):
    """Filter for where the lhs is in the RHS. The RHS must be a literal collection."""
    if isinstance(lhs_dtype, pl.Struct):
        # Assert equality field by field
        filters = []
        rhs_by_fields: Dict[str, List[Any]] = {}
        for item in rhs:
            for field in lhs_dtype.fields:
                if field.name not in rhs_by_fields:
                    rhs_by_fields[field.name] = []
                # Assuming that struct-like objects make their members accessible by attribute name or __getitem__
                try:
                    rhs_vector = getattr(item, field.name)
                except AttributeError:
                    rhs_vector = item[field.name]
                rhs_by_fields[field.name].append(rhs_vector)
        for field in lhs_dtype.fields:
            field_name = field.name
            new_lhs = lhs.struct.field(field_name)
            new_lhs_dtype = field.dtype
            new_rhs = rhs_by_fields[field_name]
            filters.append(_polars_is_in(new_lhs, new_rhs, new_lhs_dtype))
        assert len(filters) > 0, "structs with 0 fields are unsupported"
        return functools.reduce(lambda a, b: a & b, filters)
    if not isinstance(lhs_dtype, type):
        lhs_dtype = type(lhs_dtype)
    return lhs.is_in(pl.lit(pl.Series(values=rhs, dtype=lhs_dtype), allow_object=True))


def _coerce_value_to_dtype(val: Any, dtype: Union[pl.DataType, Type[pl.DataType]]):
    # For the most part, we don't need to do any manual coercion -- polars will handle that for us.
    # Only need to pay attention to enums, which could be stored as the underlying type, or as an object
    if isinstance(val, collections.abc.Iterable) and not isinstance(val, str):
        return [_coerce_value_to_dtype(x, dtype) for x in val]
    if isinstance(val, enum.Enum) and dtype != pl.Object:
        return val.value
    return val


def _convert_filter_to_pl_expr(
    f: Filter,
    df_schema: dict[str, pl.PolarsDataType],
    timestamp_feature: Optional[Feature] = None,
    now: Optional[datetime.datetime] = None,
) -> pl.Expr:
    """Convert filters to a polars expression

    Args:
        f: The filter
        df_schema: The DataFrame schema.
        timestamp_feature: The feature corresponding to the observation time
        now: The datetime to use for the current timestamp. Used to resolve relative
            timestamps in filters to absolute datetimes.

    Returns:
        A series of boolean values that can be used to select the rows where the filter is truthy
    """
    # Passing `now` in explicitly instead of using datetime.datetime.now() so that multiple filters
    # relying on relative timestamps (e.g. before, after) will have the same "now" time.
    if f.operation == "not":
        assert f.rhs is None, "not has just one side"
        assert isinstance(f.lhs, Filter), "lhs must be a filter"
        return ~_convert_filter_to_pl_expr(f.lhs, df_schema, timestamp_feature, now)
    elif f.operation == "and":
        assert isinstance(f.rhs, Filter), "rhs must be a filter"
        assert isinstance(f.lhs, Filter), "lhs must be a filter"
        return _convert_filter_to_pl_expr(f.lhs, df_schema, timestamp_feature, now) & _convert_filter_to_pl_expr(
            f.rhs, df_schema, timestamp_feature, now
        )
    elif f.operation == "or":
        assert isinstance(f.rhs, Filter), "rhs must be a filter"
        assert isinstance(f.lhs, Filter), "lhs must be a filter"
        return _convert_filter_to_pl_expr(f.lhs, df_schema, timestamp_feature, now) | _convert_filter_to_pl_expr(
            f.rhs, df_schema, timestamp_feature, now
        )

    lhs = _parse_feature_or_value(f.lhs, timestamp_feature, now)
    rhs = _parse_feature_or_value(f.rhs, timestamp_feature, now)

    lhs_dtype = None
    if isinstance(lhs, Feature):
        lhs_dtype = df_schema[str(lhs)]
        lhs = pl.col(str(lhs))

    rhs_dtype = None
    if isinstance(rhs, Feature):
        rhs_dtype = df_schema[str(rhs)]
        rhs = pl.col(str(rhs))

    if lhs_dtype is None:
        # LHS is literal. Encode it into the rhs_dtype
        assert rhs_dtype is not None
        lhs = _coerce_value_to_dtype(lhs, rhs_dtype)
    if rhs_dtype is None:
        # RHS is literal. Encode it into the lhs_dtype
        assert lhs_dtype is not None
        rhs = _coerce_value_to_dtype(rhs, lhs_dtype)

    if rhs is None:
        assert isinstance(lhs, pl.Expr)
        if f.operation == "==":
            return lhs.is_null()

        elif f.operation == "!=":
            return lhs.is_not_null()

    if f.operation in ("in", "not in"):
        assert lhs_dtype is not None
        assert isinstance(lhs, pl.Expr)
        assert isinstance(rhs, collections.abc.Iterable)
        ret = _polars_is_in(lhs, rhs, lhs_dtype)
        if f.operation == "not in":
            ret = ~ret
    elif f.operation in ("==", "!="):
        ret = _polars_is_eq(lhs, rhs, lhs_dtype, rhs_dtype)
        if f.operation == "!=":
            ret = ~ret
    elif f.operation == "!=":
        ret = lhs != rhs
    elif f.operation == ">=":
        ret = lhs >= rhs  # type: ignore
    elif f.operation == ">":
        ret = lhs > rhs  # type: ignore
    elif f.operation == "<":
        ret = lhs < rhs  # type: ignore
    elif f.operation == "<=":
        ret = lhs <= rhs  # type: ignore
    else:
        raise ValueError(f'Unknown operation "{f.operation}"')
    assert isinstance(ret, pl.Expr)
    return ret


def convert_filters_to_pl_expr(
    filters: Sequence[Filter],
    df_schema: dict[str, pl.PolarsDataType],
    timestamp_feature: Optional[Feature] = None,
    now: Optional[datetime.datetime] = None,
):
    if len(filters) == 0:
        return None
    polars_filters = (_convert_filter_to_pl_expr(f, df_schema, timestamp_feature, now) for f in filters)
    return functools.reduce(lambda a, b: a & b, polars_filters)


def _generate_empty_series_for_dtype(name: str, dtype: Union[Type[pl.DataType], pl.DataType], length: int) -> pl.Series:
    """Safely generate a series of all null values for the specified datatype.

    Unlike the ``pl.Series`` constructor, this function can handle struct dtypes.
    """
    if isinstance(dtype, pl.Struct):
        # Struct dtypes cannot be specified in the pl.Series constructor.
        # Instead, create a dataframe, then call .to_struct() on it
        # If recursing within a struct, it should have a length of zero
        data = {f.name: _generate_empty_series_for_dtype(f.name, f.dtype, length) for f in dtype.fields}
        temp_df = pl.DataFrame(data)
        return temp_df.to_struct(name)
    if isinstance(dtype, pl.List):
        assert dtype.inner is not None
        data = {name: _generate_empty_series_for_dtype(name, dtype.inner, 0)}
        temp_df = pl.DataFrame(data)
        list_of_struct_series = temp_df.select(pl.col(name).reshape((length, -1))).get_column(name)
        return list_of_struct_series
    return pl.Series(name, dtype=dtype, values=([None] * length))


# list [[], [None, None]]


def _preprocess_element(element: Any):
    if isinstance(element, enum.Enum):
        element = element.value
    # Polars has trouble with third-party timezone libraries,
    # So convert it to python's internal timezone library
    if isinstance(element, datetime.datetime) and element.tzinfo is not None:
        element = element.astimezone(datetime.timezone.utc)
    return element



import typing
from typing import Mapping, Optional, Sequence, Union

import sqlalchemy
from polars import from_arrow
from sqlalchemy import text
from sqlalchemy.sql import selectable

from chalk.clogging import chalk_logger
from chalk.features import DataFrame, Feature
from chalk.sql import BaseSQLSourceProtocol, IncrementalSettings
from chalk.sql.base.protocols import CHALK_QUERY_LOGGING, DBSessionProtocol, StringChalkQueryProtocol
from chalk.sql.integrations.chalk_query import (
    Finalizer,
    _construct_features_df,
    _construct_features_single,
    _resolve_implicit_col_feature_mappings,
)
from chalk.utils.duration import Duration


class StringChalkQuery(StringChalkQueryProtocol):
    def __init__(
        self,
        session: DBSessionProtocol,
        source: BaseSQLSourceProtocol,
        query: Union[str, selectable.Selectable],
        fields: Mapping[str, Union[Feature, str]],
        args: Optional[Mapping[str, str]],
    ):
        self._finalizer: Optional[Finalizer] = None
        self._session = session
        self._source = source
        self._original_query = query
        self._query = text(query) if isinstance(query, str) else query
        self._fields = fields
        self._args = args
        self._incremental_settings: Optional[Union[IncrementalSettings, bool]] = None
        if args is not None:
            self._query = self._query.bindparams(**args)

    def __repr__(self):
        return f"StringChalkQuery(query='{self._query}')"

    def one_or_none(self):
        self._finalizer = Finalizer.OneOrNone
        return self

    def one(self):
        self._finalizer = Finalizer.One
        return self

    def first(self):
        self._finalizer = Finalizer.First
        return self

    def all(self):
        self._finalizer = Finalizer.All
        return self

    def incremental(
        self, *, incremental_column: Optional[str] = None, lookback_period: Duration = "0s", mode: str = "row"
    ):
        if mode in {"row", "group"} and incremental_column is None:
            raise ValueError(f"incremental mode set to '{mode}' but no 'incremental_column' argument was passed.")

        if mode == "parameter" and incremental_column is not None:
            raise ValueError(
                f"incremental mode set to '{mode}' but 'incremental_column' argument was passed."
                + " Please view documentation for proper usage."
            )

        self._finalizer = Finalizer.All
        self._incremental_settings = IncrementalSettings(
            lookback_period=lookback_period, incremental_column=incremental_column, mode=mode
        )
        return self

    def execute(self):
        return self.execute_internal(expected_features=[])

    def execute_internal(self, *, expected_features: Sequence[Feature]):
        from chalk.sql import SnowflakeSourceImpl

        if CHALK_QUERY_LOGGING:
            chalk_logger.info(f"Executing query: {self._query}")

        if isinstance(self._source, SnowflakeSourceImpl) and (
            self._finalizer is None or self._finalizer == Finalizer.All
        ):
            return self._execute_internal_snowflake(expected_features=expected_features)
        else:
            return self._execute_internal_sql(expected_features=expected_features)

    def _execute_internal_snowflake(self, expected_features: Sequence[Feature]):
        # this import is safe because the only way we end up here is if we have a valid SnowflakeSource constructed,
        # which already gates this import
        import snowflake.connector

        from chalk.sql import SnowflakeSourceImpl

        source = typing.cast(SnowflakeSourceImpl, self._source)
        with snowflake.connector.connect(
            user=source.user,
            account=source.account_identifier,
            password=source.password,
            warehouse=source.warehouse,
            schema=source.schema,
            database=source.db,
        ) as con:
            cursor = con.cursor().execute(
                self._original_query.replace(":chalk_incremental_timestamp", "%(chalk_incremental_timestamp)s"),
                params=self._args,
            )

            # cursor.description is only available after execute(). If you need column metadata beforehand,
            # consider cursor.describe() https://docs.snowflake.com/en/user-guide/python-connector-example.html#retrieving-column-metadata
            col_names = [c.name for c in cursor.description]
            col_feature_mappings = _resolve_implicit_col_feature_mappings(
                # Note: snowflake by default capitalizes all table and column names, although it's
                # technically possible to disable this behavior by quoting things when creating your schema
                explicit_mappings={k.upper(): v for k, v in self._fields_normalized().items()},
                row_column_names=col_names,
                expected_features=expected_features,
            )

            arrows = cursor.fetch_arrow_all()

            if arrows is None:
                return DataFrame([])

            return DataFrame(
                from_arrow(arrows)
                .drop([col for col, f in zip(col_names, col_feature_mappings) if f is None])
                .rename({col: f.root_fqn for col, f in zip(col_names, col_feature_mappings) if f is not None})
            )

    def _execute_internal_sql(self, expected_features: Sequence[Feature]):
        cursor = self._session.execute(self._query)
        assert isinstance(cursor, sqlalchemy.engine.CursorResult)
        self._finalizer = self._finalizer or Finalizer.All

        if self._finalizer == Finalizer.All:
            rows = cursor.all()
            col_feature_mappings = _resolve_implicit_col_feature_mappings(
                explicit_mappings=self._fields_normalized(),
                expected_features=expected_features,
                row_column_names=list(cursor.keys()),
            )

            return _construct_features_df(col_feature_mappings, rows)

        row = None
        if self._finalizer == Finalizer.One:
            row = cursor.one()

        if self._finalizer == Finalizer.First:
            row = cursor.first()

        if self._finalizer == Finalizer.OneOrNone:
            row = cursor.one_or_none()

        col_feature_mappings = _resolve_implicit_col_feature_mappings(
            explicit_mappings=self._fields_normalized(),
            expected_features=expected_features,
            row_column_names=list(cursor.keys()),
        )

        # We only want to run feature_codecs for StringQuerys, not ChalkQuerys, as SqlAlchemy ORM mappings won't run in this case
        return _construct_features_single(col_feature_mappings, row, _coerce_raw_values=True)

    def _fields_normalized(self):
        return {k: Feature.from_root_fqn(v) if isinstance(v, str) else v for (k, v) in self._fields.items()}



import os
import sys
from pathlib import Path
from typing import Optional


def get_classpath(x: object) -> Path:
    filepath = sys.modules[x.__module__].__file__
    assert filepath is not None
    return Path(os.path.abspath(filepath))


def get_directory_root() -> Optional[Path]:
    current = Path(os.path.dirname(os.path.abspath("dummy.txt")))
    while True:
        if any((current / f).exists() for f in ("chalk.yaml", "chalk.yml")):
            return current
        if Path(os.path.dirname(current)) == current:
            # This is '/'
            return None
        current = current.parent
